/*
 * map.h 0.1 1998/07/17 
 * 
 * instrument mapping class implementation
 * 
 * author: Kokai Istvan <pilu@digicron.com>
 * 
 * This file contains the implememtation for read and write instrument 
 * mapping file for xmtomid converter.
 * 
 * version history:
 * 
 * 0.1 - 1998/08/23 - private version 
 * 0.2 - 2004/02/11 - modification 
 * 
 */

#include "map.h"
#include <stdlib.h>
#include <iostream>

/*
 * Create the istrument map from the mapping file fname
 */	
InstMap::InstMap(char *fname){
    
    char c;
    unsigned short ch;
    unsigned short is;
    unsigned short pr;
    signed short tr[5];
    int i;
    ifstream in;
    in.open(fname);
    if (!in) {
	std::cout << "***ERROR*** " << fname << " not found.\n";
	exit(255);
	}

    in.get(c);

	for (i=0;i<256;i++) {
		chn[i]=255;
	}
    
    while (c!=';') {
	switch (c) {
	 case '#':
	    while (c!='\n'){
		in.get(c);
	    }
	    break;
	 case 'c':
	    in >> ch;
	    in >> is;
	    in >> pr;
	    
	    chn[is]=ch;
	    prg[is]=pr;
	    
	    if (ch!=9) {
			i=0;
			in >> tr[i];
		
		
			while ((i<6)&&(tr[i]!=256)) {
			    in >> tr[++i];
			}
		
			trans[is]= (char*)malloc(i*sizeof(char));
			for (int j=0;j<i;j++){
		    	trans[is][j] = tr[j]+128;
			}
			trans[is][i]='\0';
	    }
	
	    in >> tr[0];
	    vol[is] = tr[0];
	    	    	    
	    break;
	 default:
	    ;
	}
	in.get(c);
    }
}

InstMap::~InstMap(){
	int i;
	for (i=0; i<256; i++){
		if (trans[i]!=NULL){
			free(trans[i]);
		}
	}
}
	  
/*
 * Get Midi channel for instrument is
 */
char InstMap::getMidiChannel(char is){

	return chn[is];

}

/*
 * Get Midi program for instrument is
 */
char InstMap::getMidiProgram(char is){

	return prg[is];

}

/*
 * Get Midi volume for instrument is
 */
char InstMap::getVolume(char is){

	return vol[is];

}
/*
 * Get the trans. string length
 */
char InstMap::getTranspositionLength(unsigned char is){
	if (trans[is] != NULL) {
		return strlen(trans[is]);
	} else {
		return 0;
	}
}
	
/*
 * Get the transpositions no. n for instrument is
 */
int InstMap::getTransposition(char n,char is){
	int t;
	if ((n>-1)&&(n<getTranspositionLength(is))&&(trans[is]!=NULL)) {
		t= (int)trans[is][n] +128;
		t= (t>128) ? t-256 : t;
		return t;
	} else {
		return 0;
	}

}

/*
 * Print the instrument map
 */
int InstMap::print(){

	int i,j,s;
	for (i=0; i<256; i++){

		if (chn[i]!=255) {

			std::cout << "Instrument " << i << "\n";
			std::cout << "   chn.no. " << (int)chn[i] << " program " << (int)prg[i] << "\n";
			std::cout << "   volume  " << (int)vol[i] << "\n";
			s = getTranspositionLength(i);
			std::cout << "   trans. ";
			for (j=0; j<s; j++) {
				std::cout << getTransposition(j,i) << " ";
			}
			std::cout << "\n\n";
		}
	}
}



