/*
 * midi.c 0.1 1998/08/28 
 * 
 * Midi format class implementation
 * 
 * author: Kokai Istvan <pilu@digicron.com>
 * 
 * This file contains the class implementation for read and write Midi files.
 * 
 * version history:
 * 
 * 0.1 - 1998/08/28 - private version 
 * 0.2 - 2004/02/11 - modification 
 * 
 */

#include "midi.h"

Midi::Midi(char* fn, XM* xm, InstMap* map, 
           unsigned char* CpyRighter, double tempofactor) {


	fname = fn;

	header = new MidiHeader(&outmid);

	tracks = (MidiTrack**)malloc(17*sizeof(MidiTrack));

	tracks[0] = new MidiTrack(&outmid);
	unsigned char Copyright[31] = { 0xFF, 0x02, 0x00 ,\
	'(','c',')','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0',\
	'\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0'};
	// Set the Copyright string.
	{
		int i;
		for (i = 6; i < 31 && *CpyRighter != '\0'; i++)
			Copyright[i] = *CpyRighter++;
		Copyright[2] = (unsigned char) (i - 3);
		tracks[0]->writeNextEvent(0x0,Copyright,i);
	}
	unsigned char Tempo[6] = { 0xFF, 0x51, 0x03, 0x00, 0x00, 0x00 };
	// Alter the current tempo 500000us (120 beats/minute) by the factor
	// of tempofactor.
	{
		int i = (int)(500000.0*tempofactor);
		Tempo[5]=(unsigned char)i;
		i >>= 8;
		Tempo[4]=(unsigned char)i;
		i >>= 8;
		Tempo[3]=(unsigned char)i;
		tracks[0]->writeNextEvent(0x0, Tempo, 6);
	}
	
	unsigned char EOT[3] = { 0xFF, 0x2F, 0x00 };
	pp = 6;
	
	int tempo;
	int p,r,t,o;
	int note, oldnote[32];
	int ins, oldins[32], oldprg[17] ;
	int vol;
	int eft;
	int efp;
	int vel = 64;
	int tr;
	char cc;
	char chn,prg;

	for (t=0;t<32;t++){
		oldnote[t]=NULLNOTE;
		oldins[t]=NULLINS;
	}
	for (t=0;t<17;t++){
		oldprg[t]=-1;
		lt[t]=0;
	}
	
	
	p=0;
	cc=0;
	et = -1;

	// for each patterns
	for (o=0;o<xm->header->getSongLength();o++){
	p=xm->header->order[o];
	std::cout << "pattern: " << (int)p <<"\r" ; 
	
	for (r=cc;r<xm->patterns[p]->numRows;r++) { //for each row
		et++;
		for (t=0;t<xm->header->getNumberChannels();t++) { //for each track
			
			note = xm->patterns[p]->tracks[t]->getNote(r);
			ins = xm->patterns[p]->tracks[t]->getIns(r);
			
			vol = xm->patterns[p]->tracks[t]->getVol(r);
			eft = xm->patterns[p]->tracks[t]->getEft(r);
			efp = xm->patterns[p]->tracks[t]->getEfp(r);

			if ((vol>0x0F)&&(vol<0x51)) {
				vel = vol - 0x10;
			} else {
				vel = 64;
				vol = NULLVOL;
			}
			
			if ((eft==0x0C)&&(efp<0x40)) {
			    vel = (int)((efp*efp)/64);
			} else 	if (eft==0x0F) {
			    if (efp < 0x20) {
				tempo=(int)((double)efp*80000.0*tempofactor);
				Tempo[5]=(unsigned char)tempo;
				tempo >>= 8;
				Tempo[4]=(unsigned char)tempo;
				tempo >>=8;
				Tempo[3]=(unsigned char)tempo;
				tracks[0]->writeNextEvent((et-lt[0])*pp, Tempo, 6);
				lt[0]=et;
			    } else {
				tempo=(int)(60000000.0/(double)efp*tempofactor);
				Tempo[5]=(unsigned char)tempo;
				tempo >>= 8;
				Tempo[4]=(unsigned char)tempo;
				tempo >>=8;
				Tempo[3]=(unsigned char)tempo;
				tracks[0]->writeNextEvent((et-lt[0])*pp, Tempo, 6);
				lt[0]=et;
			    }
			} else if (eft==0x0D) {
			    cc = efp;
			    r = 1000;
			} else if (eft==0x0B) {
			    o = eft;
			} else { //main else
			    cc = 0;			    
			}
			    


			if (note!=NULLNOTE) { 

				if (ins!=NULLINS) {
					chn = map->getMidiChannel(ins);
					if (ins==oldins[t]) {
						
						if (note!=NOTEOFF) { //simple note on
							if (chn!=9){
							for (tr=0;tr<map->getTranspositionLength(ins);tr++) {				
								noteoff(oldnote[t]+map->getTransposition(tr,ins),
										chn,vel);
								noteon(note+map->getTransposition(tr,ins),
								       chn,vel);
							}
							oldnote[t]=note;
							} else {
							    noteoff(oldnote[t],chn,vel);
							    noteon(map->getMidiProgram(ins),chn,vel);
							    oldnote[t]=map->getMidiProgram(ins);
							}
							

						} else { //simple note off

							for (tr=0;tr<map->getTranspositionLength(ins);tr++) {
								noteoff(oldnote[t]+map->getTransposition(tr,ins),
										chn,vel);
							}
							oldnote[t]=NULLNOTE;
						
						
						}
						

					} else { // instrument change

						prg = map->getMidiProgram(ins);
						if (prg!=oldprg[chn]) {	 // program change for chn
							prgchange(chn,prg);
							oldprg[chn]=prg;
						}
						
						if (note!=NOTEOFF) { // simple note on 

							for (tr=0;tr<map->getTranspositionLength(oldins[t]);tr++) {
								noteoff(oldnote[t]+map->getTransposition(tr,oldins[t]),
										map->getMidiChannel(oldins[t]),vel);
							}
							if (chn!=9) {
							for (tr=0;tr<map->getTranspositionLength(ins);tr++) {
								noteon(note+map->getTransposition(tr,ins),
									   chn,vel);
							}
							oldnote[t]=note;
							} else {
							    noteon(map->getMidiProgram(ins),chn,vel);
							    oldnote[t]=map->getMidiProgram(ins);
							}
							oldins[t]=ins;
							
						} else { //simple note off

							for (tr=0;tr<map->getTranspositionLength(oldins[t]);tr++) {
								noteoff(oldnote[t]+map->getTransposition(tr,oldins[t]),
										map->getMidiChannel(oldins[t]),vel);
							}
							oldnote[t]=NULLNOTE;

						}
						
					} 
				} else { // instrument is NULL

					if (oldins[t]!=NULLINS) {

						ins=oldins[t];
						chn = map->getMidiChannel(ins);
				
						if (note!=NOTEOFF) { // note on with old ins. value
							if (chn!=9) {
							for (tr=0;tr<map->getTranspositionLength(oldins[t]);tr++) {
								noteoff(oldnote[t]+map->getTransposition(tr,oldins[t]),
										map->getMidiChannel(oldins[t]),vel);
							}
							
							for (tr=0;tr<map->getTranspositionLength(ins);tr++) {
								noteon(note+map->getTransposition(tr,ins),
									   chn,vel);
							}
							oldnote[t]=note;
							} else {
							    noteoff(map->getMidiProgram(ins),chn,vel);
							    noteon(map->getMidiProgram(ins),chn,vel);
							    oldnote[t]=map->getMidiProgram(ins);
							}

						} else { // simple note off

							for (tr=0;tr<map->getTranspositionLength(oldins[t]);tr++) {
								noteoff(oldnote[t]+map->getTransposition(tr,oldins[t]),
										map->getMidiChannel(oldins[t]),vel);
							}
							oldnote[t]=NULLNOTE;

						}

					}

				}

			} else { // note is NULL

				//PENDING	
				if (vol!=NULLVOL) {

					if (oldins[t]!=NULLINS) {

						ins=oldins[t];
						chn = map->getMidiChannel(ins);
						for (tr=0;tr<map->getTranspositionLength(oldins[t]);tr++) {
							aftertouch(oldnote[t]+map->getTransposition(tr,oldins[t]),
									   chn,vel);
						}		
						
					}
				}
					

			}

		} // tracks end

	} // rows end

	} //patterns end
	for (t=0;t<17;t++) {
		if (tracks[t]!=NULL) {
			tracks[t]->writeNextEvent(et-lt[t], EOT, 3);
		}
	}
	std::cout <<"\n";
	write();
}

Midi::~Midi() {


	delete header;
	int i;
	for (i=0;i<17;i++) {
		if (tracks[i]!=NULL) {
			delete tracks[i];
		}

	}
	free(tracks);
} 

int Midi::write() {


	outmid.open(fname);
	int i;
	unsigned short k = 0;
	for (i=0;i<17;i++) {
		if (tracks[i]!=NULL) {
			k++;
		}
	}

	header->setNumTracks(k);
	header->write();
	for (i=0;i<17;i++) {
		if (tracks[i]!=NULL) {
			tracks[i]->write();
			std::cout << "Track " << i << "\r";
		}
	}
	std::cout <<"\n";
	return 0;
}

int Midi::noteoff(int note,int chn, int vel){
	if ((note!=NULLNOTE)&&(chn!=-1)){
		unsigned char d[3];
		d[0] = (char)(128+chn);
		d[1] = (char)(note);
		d[2] = (char)(vel);
		if (tracks[chn]==NULL) { 
			tracks[chn] = new MidiTrack(&outmid); 
		}
		tracks[chn]->writeNextEvent((et-lt[chn])*pp, d, 3);
		lt[chn]=et;
	}
	return 0;
}

int Midi::noteon(int note,int chn, int vel){
	if ((note!=NULLNOTE) && (chn!=-1)){
		unsigned char d[3];
		d[0] = (char)(144+chn);
		d[1] = (char)(note);
		d[2] = (char)(vel);
		if (tracks[chn]==NULL) { 
			tracks[chn] = new MidiTrack(&outmid); 
		}
		tracks[chn]->writeNextEvent((et-lt[chn])*pp, d, 3);
		lt[chn]=et;
	}
	return 0;
}

int Midi::prgchange(int chn, int prg){
	if ((chn!=9) && (chn!=-1)){
		unsigned char c[2];
		c[0] = (char)(192+chn);
		c[1] = (char)(prg);
		if (tracks[chn]==NULL) { 
			tracks[chn] = new MidiTrack(&outmid); 
		}

		tracks[chn]->writeNextEvent((et-lt[chn])*pp, c, 2);
		lt[chn]=et;
	}
	return 0;
}

int Midi::aftertouch(int note,int chn, int vel){
	if ((note!=NULLNOTE) && (chn!=-1)){
		unsigned char d[3];
		d[0] = (char)(0xA0+chn);
		d[1] = (char)(note);
		d[2] = (char)(vel);
		if (tracks[chn]==NULL) { 
			tracks[chn] = new MidiTrack(&outmid); 
		}
		tracks[chn]->writeNextEvent((et-lt[chn])*pp, d, 3);
		lt[chn]=et;
	}
	return 0;
}


